# -*- coding: utf-8 -*-
"""
@file
@brief Saves and reads a :epkg:`dataframe` into a :epkg:`zip` file.
"""
import io
import os
import zipfile
import pandas
import numpy


def to_zip(df, zipfilename, zname="df.csv", **kwargs):
    """
    Saves a :epkg:`Dataframe` into a :epkg:`zip` file.
    It can be read by @see fn to_zip.

    @param      df          :epkg:`dataframe` or :epkg:`numpy:array`
    @param      zipfilename a :epkg:`*py:zipfile:ZipFile` or a filename
    @param      zname       a filename in th zipfile
    @param      kwargs      parameters for :epkg:`pandas:to_csv` or
                            :epkg:`numpy:save`
    @return                 zipfilename

    .. exref::
        :title: Saves and reads a dataframe in a zip file
        :tag: dataframe

        This shows an example on how to save and read a
        :epkg:`pandas:dataframe` directly into a zip file.

        .. runpython::
            :showcode:

            import pandas
            from pandas_streaming.df import to_zip, read_zip

            df = pandas.DataFrame([dict(a=1, b="e"),
                                   dict(b="f", a=5.7)])

            name = "dfs.zip"
            to_zip(df, name, encoding="utf-8", index=False)
            df2 = read_zip(name, encoding="utf-8")
            print(df2)

    .. exref::
        :title: Saves and reads a numpy array in a zip file
        :tag: array

        This shows an example on how to save and read a
        :epkg:`numpy:ndarray` directly into a zip file.

        .. runpython::
            :showcode:

            import numpy
            from pandas_streaming.df import to_zip, read_zip

            arr = numpy.array([[0.5, 1.5], [0.4, 1.6]])

            name = "dfsa.zip"
            to_zip(arr, name, 'arr.npy')
            arr2 = read_zip(name, 'arr.npy')
            print(arr2)
    """
    if isinstance(df, pandas.DataFrame):
        stb = io.StringIO()
        ext = os.path.splitext(zname)[-1]
        if ext == '.npy':
            raise ValueError(
                "Extension '.npy' cannot be used to save a dataframe.")
        df.to_csv(stb, **kwargs)
    elif isinstance(df, numpy.ndarray):
        stb = io.BytesIO()
        ext = os.path.splitext(zname)[-1]
        if ext != '.npy':
            raise ValueError(  # pragma: no cover
                "Extension '.npy' is required when saving a numpy array.")
        numpy.save(stb, df, **kwargs)
    else:
        raise TypeError(  # pragma: no cover
            "Type not handled {0}".format(type(df)))
    text = stb.getvalue()

    if isinstance(zipfilename, str):
        ext = os.path.splitext(zipfilename)[-1]
        if ext != '.zip':
            raise NotImplementedError(  # pragma: no cover
                "Only zip file are implemented not '{0}'.".format(ext))
        zf = zipfile.ZipFile(zipfilename, 'w')
        close = True
    elif isinstance(zipfilename, zipfile.ZipFile):
        zf = zipfilename
        close = False
    else:
        raise TypeError(  # pragma: no cover
            "No implementation for type '{0}'".format(type(zipfilename)))

    zf.writestr(zname, text)
    if close:
        zf.close()


def read_zip(zipfilename, zname="df.csv", **kwargs):
    """
    Reads a :epkg:`dataframe` from a :epkg:`zip` file.
    It can be saved by @see fn read_zip.

    @param      zipfilename a :epkg:`*py:zipfile:ZipFile` or a filename
    @param      zname       a filename in th zipfile
    @param      kwargs      parameters for :epkg:`pandas:read_csv`
    @return                 :epkg:`pandas:dataframe` or :epkg:`numpy:array`
    """
    if isinstance(zipfilename, str):
        ext = os.path.splitext(zipfilename)[-1]
        if ext != '.zip':
            raise NotImplementedError(
                "Only zip file are implemented not '{0}'.".format(ext))
        zf = zipfile.ZipFile(zipfilename, 'r')
        close = True
    elif isinstance(zipfilename, zipfile.ZipFile):
        zf = zipfilename
        close = False
    else:
        raise TypeError(  # pragma: no cover
            "No implementation for type '{0}'".format(type(zipfilename)))

    content = zf.read(zname)
    stb = io.BytesIO(content)
    ext = os.path.splitext(zname)[-1]
    if ext == '.npy':
        df = numpy.load(stb, **kwargs)
    else:
        df = pandas.read_csv(stb, **kwargs)

    if close:
        zf.close()

    return df



# -*- coding: utf-8 -*-
"""
@file
@brief Saves and reads a :epkg:`dataframe` into a :epkg:`zip` file.
"""
import os
from io import StringIO, BytesIO
try:
    from ujson import dumps
except ImportError:  # pragma: no cover
    from json import dumps
import ijson


class JsonPerRowsStream:
    """
    Reads a :epkg:`json` streams and adds
    ``,``, ``[``, ``]`` to convert a stream containing
    one :pekg:`json` object per row into one single :epkg:`json` object.
    It only implements method *readline*.
    """

    def __init__(self, st):
        """
        @param      st      stream
        """
        self.st = st
        self.begin = True
        self.newline = False
        self.end = True

    def readline(self, size=-1):
        """
        Reads a line, adds ``,``, ``[``, ``]`` if needed.
        So the number of read characters is not recessarily
        the requested one but could be greater.
        """
        text = self.st.readline(size)
        if size == 0:
            return text
        if self.newline:
            text = ',' + text
            self.newline = False
        elif self.begin:
            text = '[' + text
            self.begin = False

        if text.endswith("\n"):
            self.newline = True
            return text
        elif len(text) == 0 or len(text) < size:
            if self.end:
                self.end = False
                return text + ']'
            else:
                return text
        else:
            return text

    def read(self, size=-1):
        """
        Reads characters, adds ``,``, ``[``, ``]`` if needed.
        So the number of read characters is not recessarily
        the requested one but could be greater.
        """
        text = self.st.read(size)
        if isinstance(text, bytes):
            cst = b"\n", b"\n,", b",", b"[", b"]"
        else:
            cst = "\n", "\n,", ",", "[", "]"
        if size == 0:
            return text
        if len(text) > 1:
            t1, t2 = text[:len(text) - 1], text[len(text) - 1:]
            t1 = t1.replace(cst[0], cst[1])
            text = t1 + t2

        if self.newline:
            text = cst[2] + text
            self.newline = False
        elif self.begin:
            text = cst[3] + text
            self.begin = False

        if text.endswith(cst[0]):
            self.newline = True
            return text
        elif len(text) == 0 or len(text) < size:
            if self.end:
                self.end = False
                return text + cst[4]
            else:
                return text
        else:
            return text

    def getvalue(self):
        """
        Returns the whole stream content.
        """
        def byline():
            line = self.readline()
            while line:
                yield line
                line = self.readline()
        return "".join(byline())


def flatten_dictionary(dico, sep="_"):
    """
    Flattens a dictionary with nested structure to a dictionary with no
    hierarchy.
    :param dico: dictionary to flatten
    :param sep: string to separate dictionary keys by
    :return: flattened dictionary

    Inspired from `flatten_json <https://github.com/amirziai/flatten/blob/master/flatten_json.py>`_.
    """
    flattened_dict = dict()

    def _flatten(obj, key):
        if obj is None:
            flattened_dict[key] = obj
        elif isinstance(obj, dict):
            for k, v in obj.items():
                if not isinstance(k, str):
                    raise TypeError(
                        "All keys must a string.")  # pragma: no cover
                k2 = k if key is None else "{0}{1}{2}".format(key, sep, k)
                _flatten(v, k2)
        elif isinstance(obj, (list, set)):
            for index, item in enumerate(obj):
                k2 = k if key is None else "{0}{1}{2}".format(key, sep, index)
                _flatten(item, k2)
        else:
            flattened_dict[key] = obj

    _flatten(dico, None)
    return flattened_dict


def enumerate_json_items(filename, encoding=None, lines=False, flatten=False, fLOG=None):
    """
    Enumerates items from a :epkg:`JSON` file or string.

    @param      filename        filename or string or stream to parse
    @param      encoding        encoding
    @param      lines           one record per row
    @param      flatten         call @see fn flatten_dictionary
    @param      fLOG            logging function
    @return                     iterator on records at first level.

    It assumes the syntax follows the format: ``[ {"id":1, ...}, {"id": 2, ...}, ...]``.
    However, if option *lines* if true, the function considers that the
    stream or file does have one record per row as follows:

        {"id":1, ...}
        {"id": 2, ...}

    .. exref::
        :title: Processes a json file by streaming.

        The module :epkg:`ijson` can read a :epkg:`JSON` file by streaming.
        This module is needed because a record can be written on multiple lines.
        This function leverages it produces the following results.

        .. runpython::
            :showcode:

            from pandas_streaming.df.dataframe_io_helpers import enumerate_json_items

            text_json = b'''
                [
                {
                    "glossary": {
                        "title": "example glossary",
                        "GlossDiv": {
                            "title": "S",
                            "GlossList": [{
                                "GlossEntry": {
                                    "ID": "SGML",
                                    "SortAs": "SGML",
                                    "GlossTerm": "Standard Generalized Markup Language",
                                    "Acronym": "SGML",
                                    "Abbrev": "ISO 8879:1986",
                                    "GlossDef": {
                                        "para": "A meta-markup language, used to create markup languages such as DocBook.",
                                        "GlossSeeAlso": ["GML", "XML"]
                                    },
                                    "GlossSee": "markup"
                                }
                            }]
                        }
                    }
                },
                {
                    "glossary": {
                        "title": "example glossary",
                        "GlossDiv": {
                            "title": "S",
                            "GlossList": {
                                "GlossEntry": [{
                                    "ID": "SGML",
                                    "SortAs": "SGML",
                                    "GlossTerm": "Standard Generalized Markup Language",
                                    "Acronym": "SGML",
                                    "Abbrev": "ISO 8879:1986",
                                    "GlossDef": {
                                        "para": "A meta-markup language, used to create markup languages such as DocBook.",
                                        "GlossSeeAlso": ["GML", "XML"]
                                    },
                                    "GlossSee": "markup"
                                }]
                            }
                        }
                    }
                }
                ]
            '''

            for item in enumerate_json_items(text_json):
                print(item)
    """
    if isinstance(filename, str):
        if "{" not in filename and os.path.exists(filename):
            with open(filename, "r", encoding=encoding) as f:
                for el in enumerate_json_items(f, encoding=encoding, lines=lines, flatten=flatten, fLOG=fLOG):
                    yield el
        else:
            st = StringIO(filename)
            for el in enumerate_json_items(st, encoding=encoding, lines=lines, flatten=flatten, fLOG=fLOG):
                yield el
    elif isinstance(filename, bytes):
        st = BytesIO(filename)
        for el in enumerate_json_items(st, encoding=encoding, lines=lines, flatten=flatten, fLOG=fLOG):
            yield el
    elif lines:
        for el in enumerate_json_items(JsonPerRowsStream(filename),
                                       encoding=encoding, lines=False, flatten=flatten, fLOG=fLOG):
            yield el
    else:
        parser = ijson.parse(filename)
        current = None
        curkey = None
        stack = []
        nbyield = 0
        for i, (_, event, value) in enumerate(parser):
            if i % 1000000 == 0 and fLOG is not None:
                fLOG(  # pragma: no cover
                    "[enumerate_json_items] i={0} yielded={1}".format(i, nbyield))
            if event == "start_array":
                if curkey is None:
                    current = []
                else:
                    if not isinstance(current, dict):
                        raise RuntimeError(  # pragma: no cover
                            "Type issue {0}".format(type(current)))
                    c = []
                    current[curkey] = c
                    current = c
                curkey = None
                stack.append(current)
            elif event == "end_array":
                stack.pop()
                if len(stack) == 0:
                    # We should be done.
                    current = None
                else:
                    current = stack[-1]
            elif event == "start_map":
                c = {}
                if curkey is None:
                    if current is None:
                        current = []
                    current.append(c)
                else:
                    current[curkey] = c  # pylint: disable=E1137
                stack.append(c)
                current = c
                curkey = None
            elif event == "end_map":
                stack.pop()
                current = stack[-1]
                if len(stack) == 1:
                    nbyield += 1
                    if flatten:
                        yield flatten_dictionary(current[-1])
                    else:
                        yield current[-1]
                    # We clear the memory.
                    current.clear()
            elif event == "map_key":
                curkey = value
            elif event in {"string", "number", "boolean"}:
                if curkey is None:
                    current.append(value)
                else:
                    current[curkey] = value  # pylint: disable=E1137
                    curkey = None
            elif event == "null":
                if curkey is None:
                    current.append(None)
                else:
                    current[curkey] = None  # pylint: disable=E1137
                    curkey = None
            else:
                raise ValueError("Unknown event '{0}'".format(
                    event))  # pragma: no cover


class JsonIterator2Stream:
    """
    Transforms an iterator on :epkg:`JSON` items
    into a stream which returns an items as a string every time
    method *read* is called.
    The iterator could be one returned by @see fn enumerate_json_items.

    .. exref::
        :title: Reshape a json file

        The function @see fn enumerate_json_items reads any
        :epkg:`json` even if every record is split over
        multiple lines. Class @see cl JsonIterator2Stream
        mocks this iterator as a stream. Each row is a single item.

        .. runpython::
            :showcode:

            from pandas_streaming.df.dataframe_io_helpers import enumerate_json_items, JsonIterator2Stream

            text_json = b'''
                [
                {
                    "glossary": {
                        "title": "example glossary",
                        "GlossDiv": {
                            "title": "S",
                            "GlossList": [{
                                "GlossEntry": {
                                    "ID": "SGML",
                                    "SortAs": "SGML",
                                    "GlossTerm": "Standard Generalized Markup Language",
                                    "Acronym": "SGML",
                                    "Abbrev": "ISO 8879:1986",
                                    "GlossDef": {
                                        "para": "A meta-markup language, used to create markup languages such as DocBook.",
                                        "GlossSeeAlso": ["GML", "XML"]
                                    },
                                    "GlossSee": "markup"
                                }
                            }]
                        }
                    }
                },
                {
                    "glossary": {
                        "title": "example glossary",
                        "GlossDiv": {
                            "title": "S",
                            "GlossList": {
                                "GlossEntry": [{
                                    "ID": "SGML",
                                    "SortAs": "SGML",
                                    "GlossTerm": "Standard Generalized Markup Language",
                                    "Acronym": "SGML",
                                    "Abbrev": "ISO 8879:1986",
                                    "GlossDef": {
                                        "para": "A meta-markup language, used to create markup languages such as DocBook.",
                                        "GlossSeeAlso": ["GML", "XML"]
                                    },
                                    "GlossSee": "markup"
                                }]
                            }
                        }
                    }
                }
                ]
            '''

            for item in JsonIterator2Stream(enumerate_json_items(text_json)):
                print(item)
    """

    def __init__(self, it, **kwargs):
        """
        @param      it      iterator
        @param      kwargs  arguments to :epkg:`*py:json:dumps`
        """
        self.it = it
        self.kwargs = kwargs

    def write(self):
        """
        The class does not write.
        """
        raise NotImplementedError()

    def read(self):
        """
        Reads the next item and returns it as a string.
        """
        try:
            value = next(self.it)
            return dumps(value, **self.kwargs)
        except StopIteration:
            return None

    def __iter__(self):
        """
        Iterate on each row.
        """
        for value in self.it:
            yield dumps(value, **self.kwargs)



# -*- coding: utf-8 -*-
"""
@file
@brief Defines a streming dataframe.
"""


class StreamingInefficientException(Exception):
    """
    Kind of operations doable with a :epkg:`pandas:DataFrame`
    but which should not be done in streaming mode.
    """

    def __init__(self, meth):
        """
        This method is inefficient in streaming mode
        and not implemented.

        @param      meth    method
        """
        Exception.__init__(
            self, "{0} should not be done in streaming mode.".format(meth))



# -*- coding: utf-8 -*-
"""
@file
@brief Defines a streaming dataframe.
"""
from io import StringIO, BytesIO
from inspect import isfunction
import numpy.random as nrandom
import pandas
from pandas.testing import assert_frame_equal
from pandas.io.json import json_normalize
from ..exc import StreamingInefficientException
from .dataframe_split import sklearn_train_test_split, sklearn_train_test_split_streaming
from .dataframe_io_helpers import enumerate_json_items, JsonIterator2Stream


class StreamingDataFrameSchemaError(Exception):
    """
    Reveals an issue with inconsistant schemas.
    """
    pass


class StreamingDataFrame:
    """
    Defines a streaming dataframe.
    The goal is to reduce the memory footprint.
    The class takes a function which creates an iterator
    on :epkg:`dataframe`. We assume this function can
    be called multiple time. As a matter of fact, the
    function is called every time the class needs to walk
    through the stream with the following loop:

    ::

        for df in self:  # self is a StreamingDataFrame
            # ...

    The constructor cannot receive an iterator otherwise
    this class would be able to walk through the data
    only once. The main reason is it is impossible to
    :epkg:`*py:pickle` (or :epkg:`dill`)
    an iterator: it cannot be replicated.
    Instead, the class takes a function which generates
    an iterator on :epkg:`DataFrame`.
    Most of the methods returns either a :epkg:`DataFrame`
    either a @see cl StreamingDataFrame. In the second case,
    methods can be chained.

    By default, the object checks that the schema remains
    the same between two chunks. This can be disabled
    by setting *check_schema=False* in the constructor.

    The user should expect the data to remain stable.
    Every loop should produce the same data. However,
    in some situations, it is more efficient not to keep
    that constraints. Draw a random @see me sample
    is one of these cases.
    """

    def __init__(self, iter_creation, check_schema=True, stable=True):
        """
        @param      iter_creation   function which creates an iterator or an instance of
                                    @see cl StreamingDataFrame
        @param      check_schema    checks that the schema is the same for every :epkg:`dataframe`
        @param      stable          indicates if the :epkg:`dataframe` remains the same whenever
                                    it is walked through
        """
        if isinstance(iter_creation, StreamingDataFrame):
            self.iter_creation = iter_creation.iter_creation
            self.stable = iter_creation.stable
        else:
            self.iter_creation = iter_creation
            self.stable = stable
        self.check_schema = check_schema

    def is_stable(self, do_check=False, n=10):
        """
        Tells if the :epkg:`dataframe` is supposed to be stable.

        @param      do_check    do not trust the value sent to the constructor
        @param      n           number of rows used to check the stability,
                                None for all rows
        @return                 boolean

        *do_check=True* means the methods checks the first
        *n* rows remains the same for two iterations.
        """
        if do_check:
            for i, (a, b) in enumerate(zip(self, self)):
                if n is not None and i >= n:
                    break
                try:
                    assert_frame_equal(a, b)
                except AssertionError:  # pragma: no cover
                    return False
            return True
        else:
            return self.stable

    def get_kwargs(self):
        """
        Returns the parameters used to call the constructor.
        """
        return dict(check_schema=self.check_schema)

    def train_test_split(self, path_or_buf=None, export_method="to_csv",
                         names=None, streaming=True, partitions=None,
                         **kwargs):
        """
        Randomly splits a :epkg:`dataframe` into smaller pieces.
        The function returns streams of file names.
        It chooses one of the options from module
        :mod:`dataframe_split <pandas_streaming.df.dataframe_split>`.

        @param  path_or_buf     a string, a list of strings or buffers, if it is a
                                string, it must contain ``{}`` like ``partition{}.txt``,
                                if None, the function returns strings.
        @param  export_method   method used to store the partitions, by default
                                :epkg:`pandas:DataFrame:to_csv`, additional parameters
                                will be given to that function
        @param  names           partitions names, by default ``('train', 'test')``
        @param  kwargs          parameters for the export function and
                                :epkg:`sklearn:model_selection:train_test_split`.
        @param  streaming       the function switches to a
                                streaming version of the algorithm.
        @param  partitions      splitting partitions
        @return                 outputs of the exports functions or two
                                @see cl StreamingDataFrame if path_or_buf is None.

        The streaming version of this algorithm is implemented by function
        @see fn sklearn_train_test_split_streaming. Its documentation
        indicates the limitation of the streaming version and gives some
        insights about the additional parameters.
        """
        if streaming:
            if partitions is not None:
                if len(partitions) != 2:
                    raise NotImplementedError(
                        "Only train and test split is allowed, *partitions* must be of length 2.")
                kwargs = kwargs.copy()
                kwargs['train_size'] = partitions[0]
                kwargs['test_size'] = partitions[1]
            return sklearn_train_test_split_streaming(self, **kwargs)
        return sklearn_train_test_split(self, path_or_buf=path_or_buf,
                                        export_method=export_method,
                                        names=names, **kwargs)

    @staticmethod
    def _process_kwargs(kwargs):
        """
        Filters out parameters for the constructor of this class.
        """
        kw = {}
        for k in {'check_schema'}:
            if k in kwargs:
                kw[k] = kwargs[k]
                del kwargs[k]
        return kw

    @staticmethod
    def read_json(*args, chunksize=100000, flatten=False, **kwargs) -> 'StreamingDataFrame':
        """
        Reads a :epkg:`json` file or buffer as an iterator
        on :epkg:`DataFrame`. The signature is the same as
        :epkg:`pandas:read_json`. The important parameter is
        *chunksize* which defines the number
        of rows to parse in a single bloc
        and it must be defined to return an iterator.
        If *lines* is True, the function falls back into
        :epkg:`pandas:read_json`, otherwise it used
        @see fn enumerate_json_items. If *lines is ``'stream'``,
        @see fn enumerate_json_items is called with parameter
        ``lines=True``.
        Parameter *flatten* uses the trick described at
        ` <https://towardsdatascience.com/flattening-json-objects-in-python-f5343c794b10>`_.
        Examples::

        .. runpython::
            :showcode:

            from io import BytesIO
            from pandas_streaming.df import StreamingDataFrame

            data = b'''{"a": 1, "b": 2}
                       {"a": 3, "b": 4}'''
            it = StreamingDataFrame.read_json(BytesIO(data), lines=True)
            dfs = list(it)
            print(dfs)

        .. runpython::
            :showcode:

            from io import BytesIO
            from pandas_streaming.df import StreamingDataFrame

            data = b'''[{"a": 1,
                         "b": 2},
                        {"a": 3,
                         "b": 4}]'''

            it = StreamingDataFrame.read_json(BytesIO(data))
            dfs = list(it)
            print(dfs)
        """
        if not isinstance(chunksize, int) or chunksize <= 0:
            raise ValueError(  # pragma: no cover
                'chunksize must be a positive integer')
        kwargs_create = StreamingDataFrame._process_kwargs(kwargs)

        if isinstance(args[0], (list, dict)):
            if flatten:
                return StreamingDataFrame.read_df(json_normalize(args[0]), **kwargs_create)
            return StreamingDataFrame.read_df(args[0], **kwargs_create)

        if kwargs.get('lines', None) == 'stream':
            del kwargs['lines']
            st = JsonIterator2Stream(enumerate_json_items(
                args[0], encoding=kwargs.get('encoding', None), lines=True, flatten=flatten))
            args = args[1:]

            if chunksize is None:
                return StreamingDataFrame(
                    lambda: pandas.read_json(
                        st, *args, chunksize=None, lines=True, **kwargs),
                    **kwargs_create)

            def fct1(st=st, args=args, chunksize=chunksize, kw=kwargs.copy()):
                for r in pandas.read_json(st, *args, chunksize=chunksize, nrows=chunksize,
                                          lines=True, **kw):
                    yield r
            return StreamingDataFrame(fct1, **kwargs_create)

        if kwargs.get('lines', False):
            if flatten:
                raise NotImplementedError(
                    "flatten==True is implemented with option lines='stream'")
            if chunksize is None:
                return StreamingDataFrame(
                    lambda: pandas.read_json(*args, chunksize=None, **kwargs),
                    **kwargs_create)

            def fct2(args=args, chunksize=chunksize, kw=kwargs.copy()):
                for r in pandas.read_json(*args, chunksize=chunksize, nrows=chunksize, **kw):
                    yield r
            return StreamingDataFrame(fct2, **kwargs_create)

        st = JsonIterator2Stream(enumerate_json_items(
            args[0], encoding=kwargs.get('encoding', None), flatten=flatten))
        args = args[1:]
        if 'lines' in kwargs:
            del kwargs['lines']

        if chunksize is None:
            return StreamingDataFrame(
                lambda: pandas.read_json(
                    st, *args, chunksize=chunksize, lines=True, **kwargs),
                **kwargs_create)

        def fct3(st=st, args=args, chunksize=chunksize, kw=kwargs.copy()):
            for r in pandas.read_json(st, *args, chunksize=chunksize, nrows=chunksize,
                                      lines=True, **kw):
                yield r
        return StreamingDataFrame(fct3, **kwargs_create)

    @staticmethod
    def read_csv(*args, **kwargs) -> 'StreamingDataFrame':
        """
        Reads a :epkg:`csv` file or buffer
        as an iterator on :epkg:`DataFrame`.
        The signature is the same as :epkg:`pandas:read_csv`.
        The important parameter is *chunksize* which defines the number
        of rows to parse in a single bloc. If not specified,
        it will be equal to 100000.
        """
        if not kwargs.get('iterator', True):
            raise ValueError("If specified, iterator must be True.")
        if not kwargs.get('chunksize', 100000):
            raise ValueError("If specified, chunksize must not be None.")
        kwargs_create = StreamingDataFrame._process_kwargs(kwargs)
        kwargs['iterator'] = True
        if 'chunksize' not in kwargs:
            kwargs['chunksize'] = 100000
        return StreamingDataFrame(lambda: pandas.read_csv(*args, **kwargs), **kwargs_create)

    @staticmethod
    def read_str(text, **kwargs) -> 'StreamingDataFrame':
        """
        Reads a :epkg:`DataFrame` as an iterator on :epkg:`DataFrame`.
        The signature is the same as :epkg:`pandas:read_csv`.
        The important parameter is *chunksize* which defines the number
        of rows to parse in a single bloc.
        """
        if not kwargs.get('iterator', True):
            raise ValueError("If specified, iterator must be True.")
        if not kwargs.get('chunksize', 100000):
            raise ValueError("If specified, chunksize must not be None.")
        kwargs_create = StreamingDataFrame._process_kwargs(kwargs)
        kwargs['iterator'] = True
        if 'chunksize' not in kwargs:
            kwargs['chunksize'] = 100000
        if isinstance(text, str):
            buffer = StringIO(text)
        else:
            buffer = BytesIO(text)
        return StreamingDataFrame(
            lambda: pandas.read_csv(buffer, **kwargs), **kwargs_create)

    @staticmethod
    def read_df(df, chunksize=None, check_schema=True) -> 'StreamingDataFrame':
        """
        Splits a :epkg:`DataFrame` into small chunks mostly for
        unit testing purposes.

        @param      df              :epkg:`DataFrame`
        @param      chunksize       number rows per chunks (// 10 by default)
        @param      check_schema    check schema between two iterations
        @return                     iterator on @see cl StreamingDataFrame
        """
        if chunksize is None:
            if hasattr(df, 'shape'):
                chunksize = df.shape[0]
            else:
                raise NotImplementedError(
                    "Cannot retrieve size to infer chunksize for type={0}".format(type(df)))

        if hasattr(df, 'shape'):
            size = df.shape[0]
        else:
            raise NotImplementedError(
                "Cannot retrieve size for type={0}".format(type(df)))

        def local_iterator():
            "local iterator"
            for i in range(0, size, chunksize):
                end = min(size, i + chunksize)
                yield df[i:end].copy()
        return StreamingDataFrame(local_iterator, check_schema=check_schema)

    def __iter__(self):
        """
        Iterator on a large file with a sliding window.
        Each windows is a :epkg:`DataFrame`.
        The method stores a copy of the initial iterator
        and restores it after the end of the iterations.
        If *check_schema* was enabled when calling the constructor,
        the method checks that every :epkg:`DataFrame`
        follows the same schema as the first chunck.

        Even with a big chunk size, it might happen
        that consecutive chunks might detect different type
        for one particular column. An error message shows up
        saying ``Column types are different after row``
        with more information about the column which failed.
        In that case, :epkg:`pandas:DataFrame.read_csv` can overwrite
        the type on one column by specifying
        ``dtype={column_name: new_type}``. It frequently happens
        when a string column has many missing values.
        """
        iters = self.iter_creation()
        sch = None
        rows = 0
        for it in iters:
            if sch is None:
                sch = (list(it.columns), list(it.dtypes))
            elif self.check_schema:
                if list(it.columns) != sch[0]:  # pylint: disable=E1136
                    raise StreamingDataFrameSchemaError(  # pragma: no cover
                        'Column names are different after row {0}\nFirst   chunk: {1}'
                        '\nCurrent chunk: {2}'.format(
                            rows, sch[0], list(it.columns)))  # pylint: disable=E1136
                if list(it.dtypes) != sch[1]:  # pylint: disable=E1136
                    errdf = pandas.DataFrame(
                        dict(names=sch[0], schema1=sch[1], schema2=list(it.dtypes)))  # pylint: disable=E1136
                    tdf = StringIO()
                    errdf['diff'] = errdf['schema2'] != errdf['schema1']
                    errdf = errdf[errdf['diff']]
                    errdf.to_csv(tdf, sep=",")
                    msg = 'Column types are different after row {0}\n{1}'
                    raise StreamingDataFrameSchemaError(
                        msg.format(rows, tdf.getvalue()))
            rows += it.shape[0]
            yield it

    def sort_values(self, *args, **kwargs):
        """
        Not implemented.
        """
        raise StreamingInefficientException(StreamingDataFrame.sort_values)

    @property
    def shape(self):
        """
        This is the kind of operations you do not want to do
        when a file is large because it goes through the whole
        stream just to get the number of rows.
        """
        nl, nc = 0, 0
        for it in self:
            nc = max(it.shape[1], nc)
            nl += it.shape[0]
        return nl, nc

    @property
    def columns(self):
        """
        See :epkg:`pandas:DataFrame:columns`.
        """
        for it in self:
            return it.columns

    @property
    def dtypes(self):
        """
        See :epkg:`pandas:DataFrame:dtypes`.
        """
        for it in self:
            return it.dtypes

    def to_csv(self, path_or_buf=None, **kwargs) -> 'StreamingDataFrame':
        """
        Saves the :epkg:`DataFrame` into string.
        See :epkg:`pandas:DataFrame.to_csv`.
        """
        if path_or_buf is None:
            st = StringIO()
            close = False
        elif isinstance(path_or_buf, str):
            st = open(path_or_buf, "w", encoding=kwargs.get('encoding'))
            close = True
        else:
            st = path_or_buf
            close = False

        for df in self:
            df.to_csv(st, **kwargs)
            kwargs['header'] = False

        if close:
            st.close()
        if isinstance(st, StringIO):
            return st.getvalue()
        return path_or_buf

    def to_dataframe(self) -> pandas.DataFrame:
        """
        Converts everything into a single :epkg:`DataFrame`.
        """
        return pandas.concat(self, axis=0)

    def to_df(self) -> pandas.DataFrame:
        """
        Converts everything into a single :epkg:`DataFrame`.
        """
        return self.to_dataframe()

    def iterrows(self):
        """
        See :epkg:`pandas:DataFrame:iterrows`.
        """
        for df in self:
            for it in df.iterrows():
                yield it

    def head(self, n=5) -> pandas.DataFrame:
        """
        Returns the first rows as a :epkg:`DataFrame`.
        """
        st = []
        total = 0
        for df in self:
            h = df.head(n=n)
            total += h.shape[0]
            st.append(h)
            if total >= n:
                break
            n -= h.shape[0]
        if len(st) == 1:
            return st[0]
        if len(st) == 0:
            return None
        return pandas.concat(st, axis=0)

    def tail(self, n=5) -> pandas.DataFrame:
        """
        Returns the last rows as a :epkg:`DataFrame`.
        The size of chunks must be greater than ``n`` to
        get ``n`` lines. This method is not efficient
        because the whole dataset must be walked through.
        """
        for df in self:
            h = df.tail(n=n)
        return h

    def where(self, *args, **kwargs) -> 'StreamingDataFrame':
        """
        Applies :epkg:`pandas:DataFrame:where`.
        *inplace* must be False.
        This function returns a @see cl StreamingDataFrame.
        """
        kwargs['inplace'] = False
        return StreamingDataFrame(
            lambda: map(lambda df: df.where(*args, **kwargs), self),
            **self.get_kwargs())

    def sample(self, reservoir=False, cache=False, **kwargs) -> 'StreamingDataFrame':
        """
        See :epkg:`pandas:DataFrame:sample`.
        Only *frac* is available, otherwise choose
        @see me reservoir_sampling.
        This function returns a @see cl StreamingDataFrame.

        @param      reservoir   use `reservoir sampling <https://en.wikipedia.org/wiki/Reservoir_sampling>`_
        @param      cache       cache the sample
        @param      kwargs      additional parameters for :epkg:`pandas:DataFrame:sample`

        If *cache* is True, the sample is cached (assuming it holds in memory).
        The second time an iterator walks through the
        """
        if reservoir or 'n' in kwargs:
            if 'frac' in kwargs:
                raise ValueError(
                    'frac cannot be specified for reservoir sampling.')
            return self._reservoir_sampling(cache=cache, n=kwargs['n'], random_state=kwargs.get('random_state'))
        if cache:
            sdf = self.sample(cache=False, **kwargs)
            df = sdf.to_df()
            return StreamingDataFrame.read_df(df, chunksize=df.shape[0])
        return StreamingDataFrame(lambda: map(lambda df: df.sample(**kwargs), self), **self.get_kwargs(), stable=False)

    def _reservoir_sampling(self, cache=True, n=1000, random_state=None) -> 'StreamingDataFrame':
        """
        Uses the `reservoir sampling <https://en.wikipedia.org/wiki/Reservoir_sampling>`_
        algorithm to draw a random sample with exactly *n* samples.

        @param      cache           cache the sample
        @param      n               number of observations to keep
        @param      random_state    sets the random_state
        @return                     @see cl StreamingDataFrame

        .. warning::
            The sample is split by chunks of size 1000.
            This parameter is not yet exposed.
        """
        if not cache:
            raise ValueError(
                "cache=False is not available for reservoir sampling.")
        indices = []
        seen = 0
        for i, df in enumerate(self):
            for ir, _ in enumerate(df.iterrows()):
                seen += 1
                if len(indices) < n:
                    indices.append((i, ir))
                else:
                    x = nrandom.random()  # pylint: disable=E1101
                    if x * n < (seen - n):
                        k = nrandom.randint(0, len(indices) - 1)
                        indices[k] = (i, ir)  # pylint: disable=E1126
        indices = set(indices)

        def reservoir_iterate(sdf, indices, chunksize):
            "iterator"
            buffer = []
            for i, df in enumerate(self):
                for ir, row in enumerate(df.iterrows()):
                    if (i, ir) in indices:
                        buffer.append(row)
                        if len(buffer) >= chunksize:
                            yield pandas.DataFrame(buffer)
                            buffer.clear()
            if len(buffer) > 0:
                yield pandas.DataFrame(buffer)

        return StreamingDataFrame(
            lambda: reservoir_iterate(sdf=self, indices=indices, chunksize=1000))

    def apply(self, *args, **kwargs) -> 'StreamingDataFrame':
        """
        Applies :epkg:`pandas:DataFrame:apply`.
        This function returns a @see cl StreamingDataFrame.
        """
        return StreamingDataFrame(
            lambda: map(lambda df: df.apply(*args, **kwargs), self),
            **self.get_kwargs())

    def applymap(self, *args, **kwargs) -> 'StreamingDataFrame':
        """
        Applies :epkg:`pandas:DataFrame:applymap`.
        This function returns a @see cl StreamingDataFrame.
        """
        return StreamingDataFrame(
            lambda: map(lambda df: df.applymap(*args, **kwargs), self),
            **self.get_kwargs())

    def merge(self, right, **kwargs) -> 'StreamingDataFrame':
        """
        Merges two @see cl StreamingDataFrame and returns @see cl StreamingDataFrame.
        *right* can be either a @see cl StreamingDataFrame or simply
        a :epkg:`pandas:DataFrame`. It calls :epkg:`pandas:DataFrame:merge` in
        a double loop, loop on *self*, loop on *right*.
        """
        if isinstance(right, pandas.DataFrame):
            return self.merge(StreamingDataFrame.read_df(right, chunksize=right.shape[0]), **kwargs)

        def iterator_merge(sdf1, sdf2, **kw):
            "iterate on dataframes"
            for df1 in sdf1:
                for df2 in sdf2:
                    df = df1.merge(df2, **kw)
                    yield df

        return StreamingDataFrame(
            lambda: iterator_merge(self, right, **kwargs), **self.get_kwargs())

    def concat(self, others, axis=0) -> 'StreamingDataFrame':
        """
        Concatenates :epkg:`dataframes`. The function ensures all :epkg:`pandas:DataFrame`
        or @see cl StreamingDataFrame share the same columns (name and type).
        Otherwise, the function fails as it cannot guess the schema without
        walking through all :epkg:`dataframes`.

        @param  others      list, enumeration, :epkg:`pandas:DataFrame`
        @return             @see cl StreamingDataFrame
        """
        if axis == 1:
            return self._concath(others)
        if axis == 0:
            return self._concatv(others)
        raise ValueError("axis must be 0 or 1")  # pragma: no cover

    def _concath(self, others):
        if not isinstance(others, list):
            others = [others]

        def iterateh(self, others):
            cols = tuple([self] + others)
            for dfs in zip(*cols):
                nrows = [_.shape[0] for _ in dfs]
                if min(nrows) != max(nrows):
                    raise RuntimeError(
                        "StreamingDataFram cannot merge DataFrame with different size or chunksize")
                yield pandas.concat(list(dfs), axis=1)

        return StreamingDataFrame(lambda: iterateh(self, others), **self.get_kwargs())

    def _concatv(self, others):

        def iterator_concat(this, lothers):
            "iterator on dataframes"
            columns = None
            dtypes = None
            for df in this:
                if columns is None:
                    columns = df.columns
                    dtypes = df.dtypes
                yield df
            for obj in lothers:
                check = True
                for i, df in enumerate(obj):
                    if check:
                        if list(columns) != list(df.columns):
                            raise ValueError(
                                "Frame others[{0}] do not have the same column names or the same order.".format(i))
                        if list(dtypes) != list(df.dtypes):
                            raise ValueError(
                                "Frame others[{0}] do not have the same column types.".format(i))
                        check = False
                    yield df

        if isinstance(others, pandas.DataFrame):
            others = [others]
        elif isinstance(others, StreamingDataFrame):
            others = [others]

        def change_type(obj):
            "change column type"
            if isinstance(obj, pandas.DataFrame):
                return StreamingDataFrame.read_df(obj, obj.shape[0])
            else:
                return obj

        others = list(map(change_type, others))
        return StreamingDataFrame(
            lambda: iterator_concat(self, others), **self.get_kwargs())

    def groupby(self, by=None, lambda_agg=None, lambda_agg_agg=None,
                in_memory=True, **kwargs) -> pandas.DataFrame:
        """
        Implements the streaming :epkg:`pandas:DataFrame:groupby`.
        We assume the result holds in memory. The out-of-memory is
        not implemented yet.

        @param      by              see :epkg:`pandas:DataFrame:groupby`
        @param      in_memory       in-memory algorithm
        @param      lambda_agg      aggregation function, *sum* by default
        @param      lambda_agg_agg  to aggregate the aggregations, *sum* by default
        @param      kwargs          additional parameters for :epkg:`pandas:DataFrame:groupby`
        @return                     :epkg:`pandas:DataFrame`

        As the input @see cl StreamingDataFrame does not necessarily hold
        in memory, the aggregation must be done at every iteration.
        There are two levels of aggregation: one to reduce every iterated
        :epkg:`dataframe`, another one to combine all the reduced :epkg:`dataframes`.
        This second one is always a **sum**.
        As a consequence, this function should not compute any *mean* or *count*,
        only *sum* because we do not know the size of each iterated
        :epkg:`dataframe`. To compute an average, sum and weights must be
        aggregated.

        Parameter *lambda_agg* is ``lambda gr: gr.sum()`` by default.
        It could also be ``lambda gr: gr.max()`` or
        ``lambda gr: gr.min()`` but not ``lambda gr: gr.mean()``
        as it would lead to incoherent results.

        .. exref::
            :title: StreamingDataFrame and groupby
            :tag: streaming

            Here is an example which shows how to write a simple *groupby*
            with :epkg:`pandas` and @see cl StreamingDataFrame.

            .. runpython::
                :showcode:

                from pandas import DataFrame
                from pandas_streaming.df import StreamingDataFrame

                df = DataFrame(dict(A=[3, 4, 3], B=[5,6, 7]))
                sdf = StreamingDataFrame.read_df(df)

                # The following:
                print(sdf.groupby("A", lambda gr: gr.sum()))

                # Is equivalent to:
                print(df.groupby("A").sum())
        """
        if not in_memory:
            raise NotImplementedError(
                "Out-of-memory group by is not implemented.")
        if lambda_agg is None:
            def lambda_agg_(gr):
                "sum"
                return gr.sum()
            lambda_agg = lambda_agg_
        if lambda_agg_agg is None:
            def lambda_agg_agg_(gr):
                "sum"
                return gr.sum()
            lambda_agg_agg = lambda_agg_agg_
        ckw = kwargs.copy()
        ckw["as_index"] = False

        agg = []
        for df in self:
            gr = df.groupby(by=by, **ckw)
            agg.append(lambda_agg(gr))
        conc = pandas.concat(agg, sort=False)
        return lambda_agg_agg(conc.groupby(by=by, **kwargs))

    def groupby_streaming(self, by=None, lambda_agg=None, lambda_agg_agg=None, in_memory=True,
                          strategy='cum', **kwargs) -> pandas.DataFrame:
        """
        Implements the streaming :epkg:`pandas:DataFrame:groupby`.
        We assume the result holds in memory. The out-of-memory is
        not implemented yet.

        @param      by              see :epkg:`pandas:DataFrame:groupby`
        @param      in_memory       in-memory algorithm
        @param      lambda_agg      aggregation function, *sum* by default
        @param      lambda_agg_agg  to aggregate the aggregations, *sum* by default
        @param      kwargs          additional parameters for :epkg:`pandas:DataFrame:groupby`
        @param      strategy        ``'cum'``, or ``'streaming'``,
                                    see below
        @return                     :epkg:`pandas:DataFrame`

        As the input @see cl StreamingDataFrame does not necessarily hold
        in memory, the aggregation must be done at every iteration.
        There are two levels of aggregation: one to reduce every iterated
        :epkg:`dataframe`, another one to combine all the reduced :epkg:`dataframes`.
        This second one is always a **sum**.
        As a consequence, this function should not compute any *mean* or *count*,
        only *sum* because we do not know the size of each iterated
        :epkg:`dataframe`. To compute an average, sum and weights must be
        aggregated.

        Parameter *lambda_agg* is ``lambda gr: gr.sum()`` by default.
        It could also be ``lambda gr: gr.max()`` or
        ``lambda gr: gr.min()`` but not ``lambda gr: gr.mean()``
        as it would lead to incoherent results.

        Parameter *strategy* allows three scenarios.
        First one if ``strategy is None`` goes through
        the whole datasets to produce a final :epkg:`DataFrame`.
        Second if ``strategy=='cum'`` returns a
        @see cl StreamingDataFrame, each iteration produces
        the current status of the *group by*. Last case,
        ``strategy=='streaming'`` produces :epkg:`DataFrame`
        which must be concatenated into a single :epkg:`DataFrame`
        and grouped again to get the results.

        .. exref::
            :title: StreamingDataFrame and groupby
            :tag: streaming

            Here is an example which shows how to write a simple *groupby*
            with :epkg:`pandas` and @see cl StreamingDataFrame.

            .. runpython::
                :showcode:

                from pandas import DataFrame
                from pandas_streaming.df import StreamingDataFrame
                from pandas_streaming.data import dummy_streaming_dataframe

                df20 = dummy_streaming_dataframe(20).to_dataframe()
                df20["key"] = df20["cint"].apply(lambda i: i % 3 == 0)
                sdf20 = StreamingDataFrame.read_df(df20, chunksize=5)
                sgr = sdf20.groupby_streaming("key", lambda gr: gr.sum(), strategy='cum', as_index=False)
                for gr in sgr:
                    print()
                    print(gr)
        """
        if not in_memory:
            raise NotImplementedError(
                "Out-of-memory group by is not implemented.")
        if lambda_agg is None:
            def lambda_agg_(gr):
                "sum"
                return gr.sum()
            lambda_agg = lambda_agg_
        if lambda_agg_agg is None:
            def lambda_agg_agg_(gr):
                "sum"
                return gr.sum()
            lambda_agg_agg = lambda_agg_agg_
        ckw = kwargs.copy()
        ckw["as_index"] = False

        if strategy == 'cum':
            def iterate_cum():
                agg = None
                for df in self:
                    gr = df.groupby(by=by, **ckw)
                    gragg = lambda_agg(gr)
                    if agg is None:
                        yield lambda_agg_agg(gragg.groupby(by=by, **kwargs))
                        agg = gragg
                    else:
                        lagg = pandas.concat([agg, gragg], sort=False)
                        yield lambda_agg_agg(lagg.groupby(by=by, **kwargs))
                        agg = lagg
            return StreamingDataFrame(lambda: iterate_cum(), **self.get_kwargs())

        if strategy == 'streaming':
            def iterate_streaming():
                for df in self:
                    gr = df.groupby(by=by, **ckw)
                    gragg = lambda_agg(gr)
                    yield lambda_agg(gragg.groupby(by=by, **kwargs))
            return StreamingDataFrame(lambda: iterate_streaming(), **self.get_kwargs())

        raise ValueError(  # pragma: no cover
            "Unknown strategy '{0}'".format(strategy))

    def ensure_dtype(self, df, dtypes):
        """
        Ensures the :epkg:`dataframe` *df* has types indicated in dtypes.
        Changes it if not.

        @param      df      dataframe
        @param      dtypes  list of types
        @return             updated?
        """
        ch = False
        cols = df.columns
        for i, (has, exp) in enumerate(zip(df.dtypes, dtypes)):
            if has != exp:
                name = cols[i]
                df[name] = df[name].astype(exp)
                ch = True
        return ch

    def __getitem__(self, *args):
        """
        Implements some of the functionalities :epkg:`pandas`
        offers for the operator ``[]``.
        """
        if len(args) != 1:
            raise NotImplementedError("Only a list of columns is supported.")
        cols = args[0]
        if not isinstance(cols, list):
            raise NotImplementedError("Only a list of columns is supported.")

        def iterate_cols(sdf):
            "iterate on columns"
            for df in sdf:
                yield df[cols]

        return StreamingDataFrame(lambda: iterate_cols(self), **self.get_kwargs())

    def add_column(self, col, value):
        """
        Implements some of the functionalities :epkg:`pandas`
        offers for the operator ``[]``.

        @param      col             new column
        @param      value           @see cl StreamingDataFrame or a lambda function
        @return                     @see cl StreamingDataFrame

        ..note::

            If value is a @see cl StreamingDataFrame,
            *chunksize* must be the same for both.

        .. exref::
            :title: Add a new column to a StreamingDataFrame
            :tag: streaming

            .. runpython::
                :showcode:

                from pandas import DataFrame
                from pandas_streaming.df import StreamingDataFrame

                df = DataFrame(data=dict(X=[4.5, 6, 7], Y=["a", "b", "c"]))
                sdf = StreamingDataFrame.read_df(df)
                sdf2 = sdf.add_column("d", lambda row: int(1))
                print(sdf2.to_dataframe())

                sdf2 = sdf.add_column("d", lambda row: int(1))
                print(sdf2.to_dataframe())

        """
        if not isinstance(col, str):
            raise NotImplementedError(
                "Only a column as a string is supported.")

        if isfunction(value):
            def iterate_fct(self, value, col):
                "iterate on rows"
                for df in self:
                    dfc = df.copy()
                    dfc.insert(dfc.shape[1], col, dfc.apply(value, axis=1))
                    yield dfc

            return StreamingDataFrame(lambda: iterate_fct(self, value, col), **self.get_kwargs())

        if isinstance(value, (pandas.Series, pandas.DataFrame, StreamingDataFrame)):
            raise NotImplementedError(
                "Unable set a new column based on a datadframe.")

        def iterate_cst(self, value, col):
            "iterate on rows"
            for df in self:
                dfc = df.copy()
                dfc[col] = value
                yield dfc

        return StreamingDataFrame(
            lambda: iterate_cst(self, value, col), **self.get_kwargs())

    def fillna(self, **kwargs):
        """
        Replaces the missing values, calls
        :epkg:`pandas:DataFrame:fillna`.

        @param      kwargs      see :epkg:`pandas:DataFrame:fillna`
        @return                 @see cl StreamingDataFrame

        .. warning::
            The function does not check what happens at the
            limit of every chunk of data. Anything but a constant value
            will probably have an inconsistent behaviour.
        """

        def iterate_na(self, **kwargs):
            "iterate on rows"
            if kwargs.get('inplace', True):
                kwargs['inplace'] = True
                for df in self:
                    df.fillna(**kwargs)
                    yield df
            else:
                for df in self:
                    yield df.fillna(**kwargs)

        return StreamingDataFrame(
            lambda: iterate_na(self, **kwargs), **self.get_kwargs())



# -*- coding: utf-8 -*-
"""
@file
@brief Helpers for dataframes.
"""
import hashlib
import struct
import warnings
import numpy
from pandas import DataFrame, Index


def hash_str(c, hash_length):
    """
    Hashes a string.

    @param      c               value to hash
    @param      hash_length     hash_length
    @return                     string
    """
    if isinstance(c, float):
        if numpy.isnan(c):
            return c
        else:
            raise ValueError("numpy.nan expected, not {0}".format(c))
    else:
        m = hashlib.sha256()
        m.update(c.encode("utf-8"))
        r = m.hexdigest()
        if len(r) >= hash_length:
            return r[:hash_length]
        return r


def hash_int(c, hash_length):
    """
    Hashes an integer into an integer.

    @param      c               value to hash
    @param      hash_length     hash_length
    @return                     int
    """
    if isinstance(c, float):
        if numpy.isnan(c):
            return c
        else:
            raise ValueError("numpy.nan expected, not {0}".format(c))
    else:
        b = struct.pack("i", c)
        m = hashlib.sha256()
        m.update(b)
        r = m.hexdigest()
        if len(r) >= hash_length:
            r = r[:hash_length]
        return int(r, 16) % (10 ** 8)


def hash_float(c, hash_length):
    """
    Hashes a float into a float.

    @param      c               value to hash
    @param      hash_length     hash_length
    @return                     int
    """
    if numpy.isnan(c):
        return c
    else:
        b = struct.pack("d", c)
        m = hashlib.sha256()
        m.update(b)
        r = m.hexdigest()
        if len(r) >= hash_length:
            r = r[:hash_length]
        i = int(r, 16) % (2 ** 53)
        return float(i)


def dataframe_hash_columns(df, cols=None, hash_length=10, inplace=False):
    """
    Hashes a set of columns in a dataframe.
    Keeps the same type. Skips missing values.

    @param      df          dataframe
    @param      cols        columns to hash or None for alls.
    @param      hash_length for strings only, length of the hash
    @param      inplace     modifies inplace
    @return                 new dataframe

    This might be useful to anonimized data before
    making it public.

    .. exref::
        :title: Hashes a set of columns in a dataframe
        :tag: dataframe

        .. runpython::
            :showcode:

            import pandas
            from pandas_streaming.df import dataframe_hash_columns
            df = pandas.DataFrame([dict(a=1, b="e", c=5.6, ind="a1", ai=1),
                                   dict(b="f", c=5.7, ind="a2", ai=2),
                                   dict(a=4, b="g", ind="a3", ai=3),
                                   dict(a=8, b="h", c=5.9, ai=4),
                                   dict(a=16, b="i", c=6.2, ind="a5", ai=5)])
            print(df)
            print('--------------')
            df2 = dataframe_hash_columns(df)
            print(df2)
    """
    if cols is None:
        cols = list(df.columns)

    if not inplace:
        df = df.copy()

    def hash_intl(c):
        "hash int"
        return hash_int(c, hash_length)

    def hash_strl(c):
        "hash string"
        return hash_str(c, hash_length)

    def hash_floatl(c):
        "hash float"
        return hash_float(c, hash_length)

    coltype = {n: t for n, t in zip(  # pylint: disable=R1721
        df.columns, df.dtypes)}  # pylint: disable=R1721
    for c in cols:
        t = coltype[c]
        if t == int:
            df[c] = df[c].apply(hash_intl)
        elif t == numpy.int64:
            df[c] = df[c].apply(lambda x: numpy.int64(hash_intl(x)))
        elif t == float:
            df[c] = df[c].apply(hash_floatl)
        elif t == object:
            df[c] = df[c].apply(hash_strl)
        else:
            raise NotImplementedError(
                "Conversion of type {0} in column '{1}' is not implemented".format(t, c))

    return df


def dataframe_unfold(df, col, new_col=None, sep=","):
    """
    One column may contain concatenated values.
    This function splits these values and multiplies the
    rows for each split value.

    @param      df      dataframe
    @param      col     column with the concatenated values (strings)
    @param      new_col new column name, if None, use default value.
    @param      sep     separator
    @return             a new dataframe

    .. exref::
        :title: Unfolds a column of a dataframe.
        :tag: dataframe

        .. runpython::
            :showcode:

            import pandas
            import numpy
            from pandas_streaming.df import dataframe_unfold

            df = pandas.DataFrame([dict(a=1, b="e,f"),
                                   dict(a=2, b="g"),
                                   dict(a=3)])
            print(df)
            df2 = dataframe_unfold(df, "b")
            print('----------')
            print(df2)

            # To fold:
            folded = df2.groupby('a').apply(lambda row: ','.join(row['b_unfold'].dropna()) \\
                                            if len(row['b_unfold'].dropna()) > 0 else numpy.nan)
            print('----------')
            print(folded)
    """
    if new_col is None:
        col_name = col + "_unfold"
    else:
        col_name = new_col
    temp_col = '__index__'
    while temp_col in df.columns:
        temp_col += "_"
    rows = []
    for i, v in enumerate(df[col]):
        if isinstance(v, str):
            spl = v.split(sep)
            for vs in spl:
                rows.append({col: v, col_name: vs, temp_col: i})
        else:
            rows.append({col: v, col_name: v, temp_col: i})
    df = df.copy()
    df[temp_col] = list(range(df.shape[0]))
    dfj = DataFrame(rows)
    res = df.merge(dfj, on=[col, temp_col])
    return res.drop(temp_col, axis=1).copy()


def dataframe_shuffle(df, random_state=None):
    """
    Shuffles a dataframe.

    @param      df              :epkg:`pandas:DataFrame`
    @param      random_state    seed
    @return                     new :epkg:`pandas:DataFrame`

    .. exref::
        :title: Shuffles the rows of a dataframe
        :tag: dataframe

        .. runpython::
            :showcode:

            import pandas
            from pandas_streaming.df import dataframe_shuffle

            df = pandas.DataFrame([dict(a=1, b="e", c=5.6, ind="a1"),
                                   dict(a=2, b="f", c=5.7, ind="a2"),
                                   dict(a=4, b="g", c=5.8, ind="a3"),
                                   dict(a=8, b="h", c=5.9, ind="a4"),
                                   dict(a=16, b="i", c=6.2, ind="a5")])
            print(df)
            print('----------')

            shuffled = dataframe_shuffle(df, random_state=0)
            print(shuffled)
    """
    if random_state is not None:
        state = numpy.random.RandomState(random_state)
        permutation = state.permutation
    else:
        permutation = numpy.random.permutation
    ori_cols = list(df.columns)
    scols = set(ori_cols)

    no_index = df.reset_index(drop=False)
    keep_cols = [_ for _ in no_index.columns if _ not in scols]
    index = no_index.index
    index = permutation(index)
    shuffled = no_index.iloc[index, :]
    res = shuffled.set_index(keep_cols)[ori_cols]
    res.index.names = df.index.names
    return res


def pandas_fillna(df, by, hasna=None, suffix=None):
    """
    Replaces the :epkg:`nan` values for something not :epkg:`nan`.
    Mostly used by @see fn pandas_groupby_nan.

    @param      df      dataframe
    @param      by      list of columns for which we need to replace nan
    @param      hasna   None or list of columns for which we need to replace NaN
    @param      suffix  use a prefix for the NaN value
    @return             list of values chosen for each column, new dataframe (new copy)
    """
    suffix = suffix if suffix else "²"
    df = df.copy()
    rep = {}
    for c in by:
        if hasna is not None and c not in hasna:
            continue
        if df[c].dtype in (str, bytes, object):
            se = set(df[c].dropna())
            val = se.pop()
            if isinstance(val, str):
                cst = suffix
                val = ""
            elif isinstance(val, bytes):
                cst = b"_"
            else:
                raise TypeError(  # pragma: no cover
                    "Unable to determine a constant for type='{0}' dtype='{1}'".format(
                        val, df[c].dtype))
            val += cst
            while val in se:
                val += suffix
            df[c].fillna(val, inplace=True)
            rep[c] = val
        else:
            dr = df[c].dropna()
            mi = abs(dr.min())
            ma = abs(dr.max())
            val = ma + mi
            if val <= ma:
                raise ValueError(  # pragma: no cover
                    "Unable to find a different value for column '{0}': min={1} max={2}"
                    "".format(val, mi, ma))
            df[c].fillna(val, inplace=True)
            rep[c] = val
    return rep, df


def pandas_groupby_nan(df, by, axis=0, as_index=False, suffix=None, nanback=True, **kwargs):
    """
    Does a *groupby* including keeping missing values (:epkg:`nan`).

    @param      df          dataframe
    @param      by          column or list of columns
    @param      axis        only 0 is allowed
    @param      as_index    should be False
    @param      suffix      None or a string
    @param      nanback     put :epkg:`nan` back in the index,
                            otherwise it leaves a replacement for :epkg:`nan`.
                            (does not work when grouping by multiple columns)
    @param      kwargs      other parameters sent to
                            `groupby <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.groupby.html>`_
    @return                 groupby results

    See `groupby and missing values <http://pandas-docs.github.io/pandas-docs-travis/groupby.html#na-and-nat-group-handling>`_.
    If no :epkg:`nan` is detected, the function falls back in regular
    :epkg:`pandas:DataFrame:groupby` which has the following
    behavior.

    .. exref::
        :title: Group a dataframe by one column including nan values
        :tag: dataframe

        The regular :epkg:`pandas:dataframe:GroupBy` of a
        :epkg:`pandas:DataFrame` removes every :epkg:`nan`
        values from the index.

        .. runpython::
            :showcode:

            from pandas import DataFrame

            data = [dict(a=2, ind="a", n=1), dict(a=2, ind="a"), dict(a=3, ind="b"), dict(a=30)]
            df = DataFrame(data)
            print(df)
            gr = df.groupby(["ind"]).sum()
            print(gr)

        Function @see fn pandas_groupby_nan modifies keeps them.

        .. runpython::
            :showcode:

            from pandas import DataFrame
            from pandas_streaming.df import pandas_groupby_nan

            data = [dict(a=2, ind="a", n=1), dict(a=2, ind="a"), dict(a=3, ind="b"), dict(a=30)]
            df = DataFrame(data)
            gr2 = pandas_groupby_nan(df, ["ind"]).sum()
            print(gr2)
    """
    if axis != 0:
        raise NotImplementedError("axis should be 0")
    if as_index:
        raise NotImplementedError("as_index must be False")
    if isinstance(by, tuple):
        raise TypeError("by should be of list not tuple")
    if not isinstance(by, list):
        by = [by]
    hasna = {}
    for b in by:
        h = df[b].isnull().values.any()
        if h:
            hasna[b] = True
    if len(hasna) > 0:
        rep, df_copy = pandas_fillna(df, by, hasna, suffix=suffix)
        res = df_copy.groupby(by, axis=axis, as_index=as_index, **kwargs)
        if len(by) == 1:
            if not nanback:
                dummy = DataFrame([{"a": "a"}])
                do = dummy.dtypes[0]
                typ = {c: t for c, t in zip(  # pylint: disable=R1721
                    df.columns, df.dtypes)}  # pylint: disable=R1721
                if typ[by[0]] != do:
                    warnings.warn(
                        "[pandas_groupby_nan] NaN value: {0}".format(rep))
                return res
            for b in by:
                fnan = rep[b]
                if fnan in res.grouper.groups:
                    res.grouper.groups[numpy.nan] = res.grouper.groups[fnan]
                    del res.grouper.groups[fnan]
                new_val = list((numpy.nan if b == fnan else b)
                               for b in res.grouper.result_index)
                res.grouper.groupings[0]._group_index = Index(new_val)
                res.grouper.groupings[0].obj[b].replace(
                    fnan, numpy.nan, inplace=True)
                if isinstance(res.grouper.groupings[0].grouper, numpy.ndarray):
                    arr = numpy.array(new_val)
                    res.grouper.groupings[0].grouper = arr
                    if hasattr(res.grouper.groupings[0], '_cache') and 'result_index' in res.grouper.groupings[0]._cache:
                        del res.grouper.groupings[0]._cache['result_index']
                else:
                    raise NotImplementedError("Not implemented for type: {0}".format(
                        type(res.grouper.groupings[0].grouper)))
                res.grouper._cache['result_index'] = res.grouper.groupings[0]._group_index
        else:
            if not nanback:
                dummy = DataFrame([{"a": "a"}])
                do = dummy.dtypes[0]
                typ = {c: t for c, t in zip(  # pylint: disable=R1721
                    df.columns, df.dtypes)}  # pylint: disable=R1721
                for b in by:
                    if typ[b] != do:
                        warnings.warn(  # pragma: no cover
                            "[pandas_groupby_nan] NaN values: {0}".format(rep))
                        break
                return res
            raise NotImplementedError(
                "Not yet implemented. Replacing pseudo nan values by real nan values is not as easy as it looks. Use nanback=False")

            # keys = list(res.grouper.groups.keys())
            # didit = False
            # mapping = {}
            # for key in keys:
            #     new_key = list(key)
            #     mod = False
            #     for k, b in enumerate(by):
            #         if b not in rep:
            #             continue
            #         fnan = rep[b]
            #         if key[k] == fnan:
            #             new_key[k] = numpy.nan
            #             mod = True
            #             didit = True
            #             mapping[fnan] = numpy.nan
            #     if mod:
            #         new_key = tuple(new_key)
            #         mapping[key] = new_key
            #         res.grouper.groups[new_key] = res.grouper.groups[key]
            #         del res.grouper.groups[key]
            # if didit:
            #     # this code deos not work
            #     vnan = numpy.nan
            #     new_index = list(mapping.get(v, v)
            #                      for v in res.grouper.result_index)
            #     names = res.grouper.result_index.names
            #     # index = MultiIndex.from_tuples(tuples=new_index, names=names)
            #     # res.grouper.result_index = index  # does not work cannot set
            #     # values for [result_index]
            #     for k in range(len(res.grouper.groupings)):
            #         grou = res.grouper.groupings[k]
            #         new_val = list(mapping.get(v, v) for v in grou)
            #         grou._group_index = Index(new_val)
            #         b = names[k]
            #         if b in rep:
            #             vv = rep[b]
            #             grou.obj[b].replace(vv, vnan, inplace=True)
            #         if isinstance(grou.grouper, numpy.ndarray):
            #             grou.grouper = numpy.array(new_val)
            #         else:
            #             raise NotImplementedError(
            #                 "Not implemented for type: {0}".format(type(grou.grouper)))
            #     del res.grouper._cache
        return res
    else:
        return df.groupby(by, axis=axis, **kwargs)


def numpy_types():
    """
    Returns the list of :epkg:`numpy` available types.

    @return     list of types
    """

    return [numpy.bool_,
            numpy.int_,
            numpy.intc,
            numpy.intp,
            numpy.int8,
            numpy.int16,
            numpy.int32,
            numpy.int64,
            numpy.uint8,
            numpy.uint16,
            numpy.uint32,
            numpy.uint64,
            numpy.float_,
            numpy.float16,
            numpy.float32,
            numpy.float64,
            numpy.complex_,
            numpy.complex64,
            numpy.complex128]



# -*- coding: utf-8 -*-
"""
@file
@brief Implements a connex split between train and test.
"""
from collections import Counter
import pandas
import numpy
from sklearn.model_selection import train_test_split
from .dataframe_helpers import dataframe_shuffle


class ImbalancedSplitException(Exception):
    """
    Raised when an imbalanced split is detected.
    """
    pass


def train_test_split_weights(df, weights=None, test_size=0.25, train_size=None,
                             shuffle=True, fail_imbalanced=0.05, random_state=None):
    """
    Splits a database in train/test given, every row
    can have a different weight.

    @param  df              :epkg:`pandas:DataFrame` or @see cl StreamingDataFrame
    @param  weights         None or weights or weights column name
    @param  test_size       ratio for the test partition (if *train_size* is not specified)
    @param  train_size      ratio for the train partition
    @param  shuffle         shuffles before the split
    @param  fail_imbalanced raises an exception if relative weights difference is higher than this value
    @param  random_state    seed for random generators
    @return                 train and test :epkg:`pandas:DataFrame`

    If the dataframe is not shuffled first, the function
    will produce two datasets which are unlikely to be randomized
    as the function tries to keep equal weights among both paths
    without using randomness.
    """
    if hasattr(df, 'iter_creation'):
        raise NotImplementedError(
            'Not implemented yet for StreamingDataFrame.')
    if isinstance(df, numpy.ndarray):
        raise NotImplementedError("Not implemented on numpy arrays.")
    if shuffle:
        df = dataframe_shuffle(df, random_state=random_state)
    if weights is None:
        if test_size == 0 or train_size == 0:
            raise ValueError(
                "test_size={0} or train_size={1} cannot be null (1)".format(test_size, train_size))
        return train_test_split(df, test_size=test_size, train_size=train_size, random_state=random_state)

    if isinstance(weights, pandas.Series):
        weights = list(weights)
    elif isinstance(weights, str):
        weights = list(df[weights])
    if len(weights) != df.shape[0]:
        raise ValueError("Dimension mismatch between weights and dataframe {0} != {1}".format(
            df.shape[0], len(weights)))

    p = (1 - test_size) if test_size else None
    if train_size is not None:
        p = train_size
        test_size = 1 - p
    if p is None or min(test_size, p) <= 0:
        raise ValueError(
            "test_size={0} or train_size={1} cannot be null (2)".format(test_size, train_size))
    ratio = test_size / p

    if random_state is None:
        randint = numpy.random.randint
    else:
        state = numpy.random.RandomState(random_state)
        randint = state.randint

    balance = 0
    train_ids = []
    test_ids = []
    test_weights = 0
    train_weights = 0
    for i in range(0, df.shape[0]):
        w = weights[i]
        if balance == 0:
            h = randint(0, 1)
            totest = h == 0
        else:
            totest = balance < 0
        if totest:
            test_ids.append(i)
            balance += w
            test_weights += w
        else:
            train_ids.append(i)
            balance -= w * ratio
            train_weights += w * ratio

    r = abs(train_weights - test_weights) / \
        (1.0 * (train_weights + test_weights))
    if r >= fail_imbalanced:
        raise ImbalancedSplitException(
            "Split is imbalanced: train_weights={0} test_weights={1} r={2}".format(train_weights, test_weights, r))

    return df.iloc[train_ids, :], df.iloc[test_ids, :]


def train_test_connex_split(df, groups, test_size=0.25, train_size=None,
                            stratify=None, hash_size=9, unique_rows=False,
                            shuffle=True, fail_imbalanced=0.05, keep_balance=None,
                            stop_if_bigger=None, return_cnx=False,
                            must_groups=None, random_state=None, fLOG=None):
    """
    This split is for a specific case where data is linked
    in many ways. Let's assume we have three ids as we have
    for online sales: *(product id, user id, card id)*.
    As we may need to compute aggregated features,
    we need every id not to be present in both train and
    test set. The function computes the connected components
    and breaks each of them in two parts for train and test.

    @param  df              :epkg:`pandas:DataFrame`
    @param  groups          columns name for the ids
    @param  test_size       ratio for the test partition (if *train_size* is not specified)
    @param  train_size      ratio for the train partition
    @param  stratify        column holding the stratification
    @param  hash_size       size of the hash to cache information about partition
    @param  unique_rows     ensures that rows are unique
    @param  shuffle         shuffles before the split
    @param  fail_imbalanced raises an exception if relative weights difference is higher than this value
    @param  stop_if_bigger  (float) stops a connected components from being
                            bigger than this ratio of elements, this should not be used
                            unless a big components emerges, the algorithm stops merging
                            but does not guarantee it returns the best cut,
                            the value should be close to 0
    @param  keep_balance    (float), if not None, does not merge connected components
                            if their relative sizes are too different, the value should be
                            close to 1
    @param  return_cnx      returns connected components as a third results
    @param  must_groups     column name for ids which must not be shared by train/test partitions
    @param  random_state    seed for random generator
    @param  fLOG            logging function
    @return                 Two @see cl StreamingDataFrame, one
                            for train, one for test.

    The list of ids must hold in memory.
    There is no streaming implementation for the ids.

    .. exref::
        :title: Splits a dataframe, keep ids in separate partitions
        :tag: dataframe

        In some data science problems, rows are not independant
        and share common value, most of the time ids. In some
        specific case, multiple ids from different columns are
        connected and must appear in the same partition.
        Testing that each id column is evenly split and do not
        appear in both sets in not enough. Connected components
        are needed.

        .. runpython::
            :showcode:

            from pandas import DataFrame
            from pandas_streaming.df import train_test_connex_split

            df = DataFrame([dict(user="UA", prod="PAA", card="C1"),
                            dict(user="UA", prod="PB", card="C1"),
                            dict(user="UB", prod="PC", card="C2"),
                            dict(user="UB", prod="PD", card="C2"),
                            dict(user="UC", prod="PAA", card="C3"),
                            dict(user="UC", prod="PF", card="C4"),
                            dict(user="UD", prod="PG", card="C5"),
                            ])

            train, test = train_test_connex_split(df, test_size=0.5,
                                                  groups=['user', 'prod', 'card'],
                                                  fail_imbalanced=0.6)
            print(train)
            print(test)

    If *return_cnx* is True, the third results contains:

    * connected components for each id
    * the dataframe with connected components as a new column

    .. runpython::
        :showcode:

        from pandas import DataFrame
        from pandas_streaming.df import train_test_connex_split

        df = DataFrame([dict(user="UA", prod="PAA", card="C1"),
                        dict(user="UA", prod="PB", card="C1"),
                        dict(user="UB", prod="PC", card="C2"),
                        dict(user="UB", prod="PD", card="C2"),
                        dict(user="UC", prod="PAA", card="C3"),
                        dict(user="UC", prod="PF", card="C4"),
                        dict(user="UD", prod="PG", card="C5"),
                        ])

        train, test, cnx = train_test_connex_split(df, test_size=0.5,
                                              groups=['user', 'prod', 'card'],
                                              fail_imbalanced=0.6, return_cnx=True)

        print(cnx[0])
        print(cnx[1])
    """
    if stratify is not None:
        raise NotImplementedError("Option stratify is not implemented.")
    if groups is None or len(groups) == 0:
        raise ValueError(  # pragma: no cover
            "groups is empty. Use regular train_test_split.")
    if hasattr(df, 'iter_creation'):
        raise NotImplementedError(
            'Not implemented yet for StreamingDataFrame.')
    if isinstance(df, numpy.ndarray):
        raise NotImplementedError("Not implemented on numpy arrays.")
    if shuffle:
        df = dataframe_shuffle(df, random_state=random_state)

    dfids = df[groups].copy()
    if must_groups is not None:
        dfids_must = df[must_groups].copy()

    name = "connex"
    while name in dfids.columns:
        name += "_"
    one = "weight"
    while one in dfids.columns:
        one += "_"

    # Connected components.
    elements = list(range(dfids.shape[0]))
    counts_cnx = {i: {i} for i in elements}
    connex = {}
    avoids_merge = {}

    def do_connex_components(dfrows, local_groups, kb, sib):
        "run connected components algorithms"
        itern = 0
        modif = 1

        while modif > 0 and itern < len(elements):
            if fLOG and df.shape[0] > 10000:
                fLOG("[train_test_connex_split] iteration={0}-#nb connect={1} - modif={2}".format(
                    iter, len(set(elements)), modif))
            modif = 0
            itern += 1
            for i, row in enumerate(dfrows.itertuples(index=False, name=None)):
                vals = [val for val in zip(local_groups, row) if not isinstance(
                    val[1], float) or not numpy.isnan(val[1])]

                c = elements[i]

                for val in vals:
                    if val not in connex:
                        connex[val] = c
                        modif += 1

                set_c = set(connex[val] for val in vals)
                set_c.add(c)
                new_c = min(set_c)

                add_pair_c = []
                for c in set_c:
                    if c == new_c or (new_c, c) in avoids_merge:
                        continue
                    if kb is not None:
                        maxi = min(len(counts_cnx[new_c]), len(counts_cnx[c]))
                        if maxi > 5:
                            diff = len(counts_cnx[new_c]) + \
                                len(counts_cnx[c]) - maxi
                            r = diff / float(maxi)
                            if r > kb:
                                if fLOG:  # pragma: no cover
                                    fLOG('[train_test_connex_split]    balance r={0:0.00000}>{1:0.00}, #[{2}]={3}, #[{4}]={5}'.format(
                                        r, kb, new_c, len(counts_cnx[new_c]), c, len(counts_cnx[c])))
                                continue

                    if sib is not None:
                        r = (len(counts_cnx[new_c]) +
                             len(counts_cnx[c])) / float(len(elements))
                        if r > sib:
                            if fLOG:  # pragma: no cover
                                fLOG('[train_test_connex_split]    no merge r={0:0.00000}>{1:0.00}, #[{2}]={3}, #[{4}]={5}'.format(
                                    r, sib, new_c, len(counts_cnx[new_c]), c, len(counts_cnx[c])))
                            avoids_merge[new_c, c] = i
                            continue

                    add_pair_c.append(c)

                if len(add_pair_c) > 0:
                    for c in add_pair_c:
                        modif += len(counts_cnx[c])
                        for ii in counts_cnx[c]:
                            elements[ii] = new_c
                        counts_cnx[new_c] = counts_cnx[new_c].union(
                            counts_cnx[c])
                        counts_cnx[c] = set()

                        keys = list(vals)
                        for val in keys:
                            if connex[val] == c:
                                connex[val] = new_c
                                modif += 1

    if must_groups:
        do_connex_components(dfids_must, must_groups, None, None)
    do_connex_components(dfids, groups, keep_balance, stop_if_bigger)

    # final
    dfids[name] = elements
    dfids[one] = 1
    grsum = dfids[[name, one]].groupby(name, as_index=False).sum()
    if fLOG:
        for g in groups:
            fLOG("[train_test_connex_split]     #nb in '{0}': {1}".format(
                g, len(set(dfids[g]))))
        fLOG(
            "[train_test_connex_split] #connex {0}/{1}".format(grsum.shape[0], dfids.shape[0]))
    if grsum.shape[0] <= 1:
        raise ValueError(  # pragma: no cover
            "Every element is in the same connected components.")

    # Statistics: top connected components
    if fLOG:
        # Global statistics
        counts = Counter(elements)
        cl = [(v, k) for k, v in counts.items()]
        cum = 0
        maxc = None
        fLOG("[train_test_connex_split] number of connected components: {0}".format(
            len(set(elements))))
        for i, (v, k) in enumerate(sorted(cl, reverse=True)):
            if i == 0:
                maxc = k, v
            if i >= 10:
                break
            cum += v
            fLOG("[train_test_connex_split]     c={0} #elements={1} cumulated={2}/{3}".format(
                k, v, cum, len(elements)))

        # Most important component
        fLOG(
            '[train_test_connex_split] first row of the biggest component {0}'.format(maxc))
        tdf = dfids[dfids[name] == maxc[0]]
        fLOG('[train_test_connex_split] \n{0}'.format(tdf.head(n=10)))

    # Splits.
    train, test = train_test_split_weights(grsum, weights=one, test_size=test_size,
                                           train_size=train_size, shuffle=shuffle,
                                           fail_imbalanced=fail_imbalanced,
                                           random_state=random_state)
    train.drop(one, inplace=True, axis=1)
    test.drop(one, inplace=True, axis=1)

    # We compute the final dataframe.
    def double_merge(d):
        "merge twice"
        merge1 = dfids.merge(d, left_on=name, right_on=name)
        merge2 = df.merge(merge1, left_on=groups, right_on=groups)
        return merge2

    train_f = double_merge(train)
    test_f = double_merge(test)
    if return_cnx:
        return train_f, test_f, (connex, dfids)
    else:
        return train_f, test_f


def train_test_apart_stratify(df, group, test_size=0.25, train_size=None,
                              stratify=None, force=False, random_state=None, fLOG=None):
    """
    This split is for a specific case where data is linked
    in one way. Let's assume we have two ids as we have
    for online sales: *(product id, category id)*.
    A product can have multiple categories. We need to have
    distinct products on train and test but common categories
    on both sides.

    @param  df              :epkg:`pandas:DataFrame`
    @param  group           columns name for the ids
    @param  test_size       ratio for the test partition (if *train_size* is not specified)
    @param  train_size      ratio for the train partition
    @param  stratify        column holding the stratification
    @param  force           if True, tries to get at least one example on the test side
                            for each value of the column *stratify*
    @param  random_state    seed for random generators
    @param  fLOG            logging function
    @return                 Two @see cl StreamingDataFrame, one
                            for train, one for test.

    .. index:: multi-label

    The list of ids must hold in memory.
    There is no streaming implementation for the ids.
    This split was implemented for a case of a multi-label
    classification. A category (*stratify*) is not exclusive
    and an observation can be assigned to multiple
    categories. In that particular case, the method
    `train_test_split <http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html>`_
    can not directly be used.

    .. runpython::
        :showcode:

        import pandas
        from pandas_streaming.df import train_test_apart_stratify

        df = pandas.DataFrame([dict(a=1, b="e"),
                               dict(a=1, b="f"),
                               dict(a=2, b="e"),
                               dict(a=2, b="f")])

        train, test = train_test_apart_stratify(
            df, group="a", stratify="b", test_size=0.5)
        print(train)
        print('-----------')
        print(test)
    """
    if stratify is None:
        raise ValueError("stratify must be specified.")
    if group is None:
        raise ValueError("group must be specified.")
    if hasattr(df, 'iter_creation'):
        raise NotImplementedError(
            'Not implemented yet for StreamingDataFrame.')
    if isinstance(df, numpy.ndarray):
        raise NotImplementedError("Not implemented on numpy arrays.")

    p = (1 - test_size) if test_size else None
    if train_size is not None:
        p = train_size
    test_size = 1 - p
    if p is None or min(test_size, p) <= 0:
        raise ValueError(  # pragma: no cover
            "test_size={0} or train_size={1} cannot be null".format(test_size, train_size))

    couples = df[[group, stratify]].itertuples(name=None, index=False)
    hist = Counter(df[stratify])
    sorted_hist = [(v, k) for k, v in hist.items()]
    sorted_hist.sort()
    ids = {c: set() for c in hist}

    for g, s in couples:
        ids[s].add(g)

    if random_state is None:
        permutation = numpy.random.permutation
    else:
        state = numpy.random.RandomState(random_state)
        permutation = state.permutation

    split = {}
    for _, k in sorted_hist:
        not_assigned = [c for c in ids[k] if c not in split]
        if len(not_assigned) == 0:
            continue
        assigned = [c for c in ids[k] if c in split]
        nb_test = sum(split[c] for c in assigned)
        expected = min(len(ids[k]), int(
            test_size * len(ids[k]) + 0.5)) - nb_test
        if force and expected == 0 and nb_test == 0:
            nb_train = len(assigned) - nb_test
            if nb_train > 0 or len(not_assigned) > 1:
                expected = min(1, len(not_assigned))
        if expected > 0:
            permutation(not_assigned)
            for e in not_assigned[:expected]:
                split[e] = 1
            for e in not_assigned[expected:]:
                split[e] = 0
        else:
            for c in not_assigned:
                split[c] = 0

    train_set = set(k for k, v in split.items() if v == 0)
    test_set = set(k for k, v in split.items() if v == 1)
    train_df = df[df[group].isin(train_set)]
    test_df = df[df[group].isin(test_set)]
    return train_df, test_df



# -*- coding: utf-8 -*-
"""
@file
@brief Dummy datasets.
"""
from pandas import DataFrame
from ..df import StreamingDataFrame


def dummy_streaming_dataframe(n, chunksize=10, asfloat=False, **cols):
    """
    Returns a dummy streaming dataframe
    mostly for unit test purposes.

    @param      n           number of rows
    @param      chunksize   chunk size
    @param      asfloat     use random float and not random int
    @param      cols        additional columns
    @return                 a @see cl StreamingDataFrame
    """
    if asfloat:
        df = DataFrame(dict(cfloat=[_ + 0.1 for _ in range(0, n)], cstr=[
                       "s{0}".format(i) for i in range(0, n)]))
    else:
        df = DataFrame(dict(cint=list(range(0, n)), cstr=[
                       "s{0}".format(i) for i in range(0, n)]))
    for k, v in cols.items():
        df[k] = v
    return StreamingDataFrame.read_df(df, chunksize=chunksize)



"""
@file
@brief Shortcuts to *df*.
"""

from .connex_split import train_test_split_weights, train_test_connex_split, train_test_apart_stratify
from .dataframe import StreamingDataFrame
from .dataframe_helpers import dataframe_hash_columns, dataframe_unfold, dataframe_shuffle
from .dataframe_helpers import pandas_groupby_nan, numpy_types
from .dataframe_io import to_zip, read_zip



# -*- coding: utf-8 -*-
"""
@file
@brief Implements different methods to split a dataframe.
"""
import hashlib
import pickle
import random
import warnings
from io import StringIO
import pandas


def sklearn_train_test_split(self, path_or_buf=None, export_method="to_csv",
                             names=None, **kwargs):
    """
    Randomly splits a dataframe into smaller pieces.
    The function returns streams of file names.
    The function relies on :epkg:`sklearn:model_selection:train_test_split`.
    It does not handle stratified version of it.

    @param  self            @see cl StreamingDataFrame
    @param  path_or_buf     a string, a list of strings or buffers, if it is a
                            string, it must contain ``{}`` like ``partition{}.txt``
    @param  export_method   method used to store the partitions, by default
                            :epkg:`pandas:DataFrame:to_csv`
    @param  names           partitions names, by default ``('train', 'test')``
    @param  kwargs          parameters for the export function and
                            :epkg:`sklearn:model_selection:train_test_split`.
    @return                 outputs of the exports functions

    The function cannot return two iterators or two
    @see cl StreamingDataFrame because running through one
    means running through the other. We can assume both
    splits do not hold in memory and we cannot run through
    the same iterator again as random draws would be different.
    We need to store the results into files or buffers.

    .. warning::
        The method *export_method* must write the data in
        mode *append* and allows stream.
    """
    if kwargs.get("stratify") is not None:
        raise NotImplementedError(
            "No implementation yet for the stratified version.")
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ImportWarning)
        from sklearn.model_selection import train_test_split  # pylint: disable=C0415

    opts = ['test_size', 'train_size',
            'random_state', 'shuffle', 'stratify']
    split_ops = {}
    for o in opts:
        if o in kwargs:
            split_ops[o] = kwargs[o]
            del kwargs[o]

    exportf_ = getattr(pandas.DataFrame, export_method)
    if export_method == 'to_csv' and 'mode' not in kwargs:
        exportf = lambda *a, **kw: exportf_(*a, mode='a', **kw)
    else:
        exportf = exportf_

    if isinstance(path_or_buf, str):
        if "{}" not in path_or_buf:
            raise ValueError(
                "path_or_buf must contain {} to insert the partition name")
        if names is None:
            names = ['train', 'test']
        elif len(names) != len(path_or_buf):
            raise ValueError(  # pragma: no cover
                'names and path_or_buf must have the same length')
        path_or_buf = [path_or_buf.format(n) for n in names]
    elif path_or_buf is None:
        path_or_buf = [None, None]
    else:
        if not isinstance(path_or_buf, list):
            raise TypeError(  # pragma: no cover
                'path_or_buf must be a list or a string')

    bufs = []
    close = []
    for p in path_or_buf:
        if p is None:
            st = StringIO()
            cl = False
        elif isinstance(p, str):
            st = open(p, "w", encoding=kwargs.get('encoding'))
            cl = True
        else:
            st = p
            cl = False
        bufs.append(st)
        close.append(cl)

    for df in self:
        train, test = train_test_split(df, **split_ops)
        exportf(train, bufs[0], **kwargs)
        exportf(test, bufs[1], **kwargs)
        kwargs['header'] = False

    for b, c in zip(bufs, close):
        if c:
            b.close()
    return [st.getvalue() if isinstance(st, StringIO) else p for st, p in zip(bufs, path_or_buf)]


def sklearn_train_test_split_streaming(self, test_size=0.25, train_size=None,
                                       stratify=None, hash_size=9, unique_rows=False):
    """
    Randomly splits a dataframe into smaller pieces.
    The function returns streams of file names.
    The function relies on :epkg:`sklearn:model_selection:train_test_split`.
    It handles the stratified version of it.

    @param  self            @see cl StreamingDataFrame
    @param  test_size       ratio for the test partition (if *train_size* is not specified)
    @param  train_size      ratio for the train partition
    @param  stratify        column holding the stratification
    @param  hash_size       size of the hash to cache information about partition
    @param  unique_rows     ensures that rows are unique
    @return                 Two @see cl StreamingDataFrame, one
                            for train, one for test.

    The function returns two iterators or two
    @see cl StreamingDataFrame. It
    tries to do everything without writing anything on disk
    but it requires to store the repartition somehow.
    This function hashes every row and maps the hash with a part
    (train or test). This cache must hold in memory otherwise the
    function fails. The two returned iterators must not be used
    for the first time in the same time. The first time is used to
    build the cache. The function changes the order of rows if
    the parameter *stratify* is not null. The cache has a side effect:
    every exact same row will be put in the same partition.
    If that is not what you want, you should add an index column
    or a random one.
    """
    p = (1 - test_size) if test_size else None
    if train_size is not None:
        p = train_size
    n = 2 * max(1 / p, 1 / (1 - p))  # changement

    static_schema = []

    def iterator_rows():
        "iterates on rows"
        counts = {}
        memory = {}
        pos_col = None
        for df in self:
            if pos_col is None:
                static_schema.append(list(df.columns))
                static_schema.append(list(df.dtypes))
                static_schema.append(df.shape[0])
                if stratify is not None:
                    pos_col = list(df.columns).index(stratify)
                else:
                    pos_col = -1

            for obs in df.itertuples(index=False, name=None):
                strat = 0 if stratify is None else obs[pos_col]
                if strat not in memory:
                    memory[strat] = []
                memory[strat].append(obs)

                for k in memory:
                    v = memory[k]
                    if len(v) >= n + random.randint(0, 10):  # changement
                        vr = list(range(len(v)))
                        # on permute aléatoirement
                        random.shuffle(vr)
                        if (0, k) in counts:
                            tt = counts[1, k] + counts[0, k]
                            delta = - int(counts[0, k] - tt * p + 0.5)
                        else:
                            delta = 0
                        i = int(len(v) * p + 0.5)
                        i += delta
                        i = max(0, min(len(v), i))
                        one = set(vr[:i])
                        for d, obs_ in enumerate(v):
                            yield obs_, 0 if d in one else 1
                        if (0, k) not in counts:
                            counts[0, k] = i
                            counts[1, k] = len(v) - i
                        else:
                            counts[0, k] += i
                            counts[1, k] += len(v) - i
                        # on efface de la mémoire les informations produites
                        memory[k].clear()

        # Lorsqu'on a fini, il faut tout de même répartir les
        # observations stockées.
        for k in memory:
            v = memory[k]
            vr = list(range(len(v)))
            # on permute aléatoirement
            random.shuffle(vr)
            if (0, k) in counts:
                tt = counts[1, k] + counts[0, k]
                delta = - int(counts[0, k] - tt * p + 0.5)
            else:
                delta = 0
            i = int(len(v) * p + 0.5)
            i += delta
            i = max(0, min(len(v), i))
            one = set(vr[:i])
            for d, obs in enumerate(v):
                yield obs, 0 if d in one else 1
            if (0, k) not in counts:
                counts[0, k] = i
                counts[1, k] = len(v) - i
            else:
                counts[0, k] += i
                counts[1, k] += len(v) - i

    def h11(w):
        "pickle and hash"
        b = pickle.dumps(w)
        return hashlib.md5(b).hexdigest()[:hash_size]

    # We store the repartition in a cache.
    cache = {}

    def iterator_internal(part_requested):
        "internal iterator on dataframes"
        iy = 0
        accumul = []
        if len(cache) == 0:
            for obs, part in iterator_rows():
                h = h11(obs)
                if unique_rows and h in cache:
                    raise ValueError(  # pragma: no cover
                        "A row or at least its hash is already cached. "
                        "Increase hash_size or check for duplicates "
                        "('{0}')\n{1}.".format(h, obs))
                if h not in cache:
                    cache[h] = part
                else:
                    part = cache[h]
                if part == part_requested:
                    accumul.append(obs)
                    if len(accumul) >= static_schema[2]:
                        dfo = pandas.DataFrame(
                            accumul, columns=static_schema[0])
                        self.ensure_dtype(dfo, static_schema[1])
                        iy += dfo.shape[0]
                        accumul.clear()
                        yield dfo
        else:
            for df in self:
                for obs in df.itertuples(index=False, name=None):
                    h = h11(obs)
                    part = cache.get(h)
                    if part is None:
                        raise ValueError(  # pragma: no cover
                            "Second iteration. A row was never met in the first one\n{0}".format(obs))
                    if part == part_requested:
                        accumul.append(obs)
                        if len(accumul) >= static_schema[2]:
                            dfo = pandas.DataFrame(
                                accumul, columns=static_schema[0])
                            self.ensure_dtype(dfo, static_schema[1])
                            iy += dfo.shape[0]
                            accumul.clear()
                            yield dfo
        if len(accumul) > 0:
            dfo = pandas.DataFrame(accumul, columns=static_schema[0])
            self.ensure_dtype(dfo, static_schema[1])
            iy += dfo.shape[0]
            yield dfo

    return (self.__class__(lambda: iterator_internal(0)),
            self.__class__(lambda: iterator_internal(1)))
